// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef INTERPOLATION_H
#define INTERPOLATION_H 
#include<cmath>
#include<iostream>
#include "interpolation.h"
#include "utility.h"
#include<vector>

/****************************************************************
*
*
*
*****************************************************************/

template<class T>
class Interpolation
{
public:
  
    Interpolation( vector<vector<float> > &image)
    {
      
        _image = image;
        _nx1 = image.size() - 1;
        _ny1 = image[0].size() - 1;
        _xl = 0.5;
        _yl = 0.5;
        _xu = float(_nx1) + 0.5;
        _yu = float(_ny1) + 0.5;

      
    }

    /************************************************************************
    *
    *
    ************************************************************************/ 
        
    inline T BilinearInterpolation(T x, T y)
    {
 
        if (x < _xl || x > _xu || y < _yl || y > _yu)
        {

            return -1.0;    

        }

	/*****************************************************************
	*
	* Find the nearest x value to the left of the interpolation point
        *
        ******************************************************************/

        int j0 = max(0, int(x) - 1);
        j0 = min(_nx1, j0);

	/*****************************************************************
	*
	* Find the nearest y value to the bottom of the interpolation point
        *
        ******************************************************************/


      
        int k0 = max(0, int(y) - 1);
        k0 = min(_ny1, k0);


        
        int j1 = j0 + 1;
        int k1 = k0 + 1;
        double t = 0.0;
        if (j1 > _nx1)
        { 
            t = 0.0;
	    j1 = _nx1;
        } 
        else
        {         
            t = max(0.0, x - double(j0 + 1));
        }
        double u = 0.0;
        if (k1 > _ny1) 
        {
            k1 = _ny1;
	    u = 0.0;
        }
        else 
        { 
	    u = max(0.0, y - double(k0 + 1));
        }
        double weight1 = 1.0, weight2 = 1.0, weight3 = 1.0, weight4 = 1.0;
        double sumWeight = 4.0;
        bool okay1 = true, okay2 = true, okay3 = true, okay4 = true;


        if( isNull( _image[j0][k0]) || _image[j0][k0] < 0.0) 
	{         
	    okay1 = false;
            weight1 = 0.0;
            sumWeight -= 1.0;
            
        }
        if( isNull( _image[j1][k0])  || _image[j1][k0] < 0.0  ) 
        {
	    okay2 = false;
            weight2 = 0.0;
            sumWeight -= 1.0;
        } 
        if( isNull( _image[j1][k1]) || _image[j1][k1] < 0.0 ) 
        {
	    okay3 = false;
            weight3 = 0.0;
            sumWeight -= 1.0;
        }
        if( isNull( _image[j0][k1]) || _image[j0][k1] < 0.0 ) 
	{   
	    okay4 = false;     
            weight4 = 0.0;
            sumWeight -= 1.0;
        }   

        double value;
        if(sumWeight > 3.5)
        {   
	


            value =  (1.0 - t) * (1.0 - u) * (double)_image[j0][k0] + 
                   t * (1.0 - u) * (double)_image[j1][k0] + 
                   t * u * (double)_image[j1][k1] + 
	           (1.0  - t) * u * (double)_image[j0][k1] ;
        }
        else if( (okay1 && okay3) || (okay2 && okay4))
	
        {   

	    /************************************************************
	    * okay1 corresponds to weight1
	    * okay2 corresponds to weight2
	    * okay3 corresponds to weight3
	    * okay4 corresponds to weight4
            *
            ************************************************************/

	  
	    sumWeight = 0.0;
	    if(okay1) 
	    {
	        weight1 = (1.0 - t) * (1.0 - u) * weight1;
                sumWeight = weight1;
            } 
            if(okay2)
	    { 
                weight2 = t * (1.0 - u) * weight2;
                sumWeight += weight2;
            }
            if(okay3)
	    { 
                weight3 = t * u;
                sumWeight += weight3;
            }
            if(okay4)
	    {
                weight4 = (1.0 - t) * u * weight4;
                sumWeight += weight4;
            }
            value = ( (double)_image[j0][k0]  * weight1 + 
                        (double)_image[j1][k0]  * weight2 + 
                        (double)_image[j1][k1]  * weight3 + 
	                (double)_image[j0][k1]  * weight4 ) / 
                   sumWeight ;
        }
        else
	{
	    value = -1.0;

        }


	//	        if(value > 0.0 && value < 700.0)
	//	{

	//   cout << "x=" << x << " y=" << y << " interpolated value=" << value << endl;
	//  cout << "j=" << j0 << " k=" << k0 << " image=" << _image[j0][k0] << " weight1=" << weight1 << endl;
	//   cout << "j=" << j1 << " k=" << k0 << " image=" << _image[j1][k0] << " weight2=" << weight2 << endl;
	//   cout << "j=" << j1 << " k=" << k1 << " image=" << _image[j1][k1] << " weight3=" << weight3 << endl;
	//   cout << "j=" << j0 << " k=" << k1 << " image=" << _image[j0][k1] << " weight4=" << weight4 << endl;
	//   cout << "Interpolated value=" << value << " sumweight=" << sumWeight << endl;
	//  exit(0);
	//	 }

        return (T)value;

    }


private:
    int _nx1, _ny1;
    double _xl, _yl;
    double _xu, _yu;
    
    vector<vector<float> > _image;
    
    

};
/*********************************************************************************
*
* Specialization for integer16
*
*********************************************************************************/
template<>
class Interpolation<int16>
{
public:
  
    Interpolation( vector<vector<int16> > &image)
    {
      
        _image = image;
        _nx1 = image.size() - 1;
        _ny1 = image[0].size() - 1;
        _xl = 0.5;
        _yl = 0.5;
        _xu = float(_nx1) + 0.5;
        _yu = float(_ny1) + 0.5;

 
    }
    /****************************************************************************
    *
    *
    *
    ****************************************************************************/
   
    inline int16 BilinearInterpolation(double x, double y)
    {


    
        if (x <= _xl || x >= _xu || y <= _yl || y >= _yu)
        {

            return (int16)-1;    

        }

        int j0 = max(0, int(x) - 1);
        j0 = min(_nx1, j0);      
        int k0 = max(0, int(y) - 1);       
        int j1 = min(_nx1, j0 + 1);
        int k1 = min(_ny1, k0 + 1);


        vector<int16> pixels;

        if( !isNull( _image[j0][k0]) && _image[j0][k0] > 0 ) pixels.push_back( _image[j0][k0] );
        if( !isNull( _image[j1][k0]) && _image[j1][k0] > 0 ) pixels.push_back( _image[j1][k0] );
        if( !isNull( _image[j1][k1]) && _image[j1][k1] > 0 ) pixels.push_back( _image[j1][k1] );
        if( !isNull( _image[j0][k1]) && _image[j0][k1] > 0 ) pixels.push_back( _image[j0][k1] );

        if(pixels.size() == 0)
        {
            return (int16)-1;
        }
        else
        {   

            int16 value = 0;
            for(unsigned j = 0; j < pixels.size(); j++)
	    { 
                //for(int i = 0; i < 10; i++) 
                for(int i = 0; i < 12; i++)  // Changed for the Jupiter patch
                {
	            if(bitSet( pixels[j], i ) )
	            {
	                setBit( value, i );
	        
	            } 
      
                }
            }
            return value;

        }      

    };


private:
    int _nx1, _ny1;
    double _xl, _yl;
    double _xu, _yu;
    
    vector<vector<int16> > _image;
    
   
};


#endif




