// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "astrometrydefs.h"
#include"omwcs.h"
#include <iostream>

#include <iomanip>
#include "Dal.h"
#include <errstr.h>
#include "utility.h"
#include <f90cxx.h>
#include <cmath>
#include <cstdlib>
#include"param.h"
//gcc4.8.2 portability changes
#include <cstdio>

/**************************************************************
*
*
*
***************************************************************/
static double d2pi= (double)6.283185307179586476925286766559;
static double tiny = (double)1e-6;


/***************************************************************
*
*
*
*****************************************************************/

inline double dranrm(double angle)
{

  double dran = fmod(angle, d2pi);
  if ( dran < 0.0) dran += d2pi;
  return dran;
}

/******************************************************************
*
* Converts tangent-plane coordinates xi, eta to celestial coordinates
* ra and dec
*
*****************************************************************/
inline void tangentPlaneToCelestial(WcsData1 &wcs, double xi, double eta, double &ra, double &dec)
{
  
    double denom = wcs._cdecz - eta * wcs._sdecz;

    ra = dranrm(atan2(xi, denom) + wcs.crval1);
    dec = atan2(wcs._sdecz + eta * wcs._cdecz, sqrt(xi * xi + denom * denom));

};
/**************************************************************
*
*
*
*
***************************************************************/
inline void celestialToTangentPlane(WcsData1 & wcs, double ra, double dec, double &x, double &y)
{
    double sdec = sin(dec);
    double cdec = cos(dec);

    double radif = ra - wcs.crval1;
    double sradif = sin(radif);
    double cradif = cos(radif);

    /********************************************************
    *
    *  Reciprocal of star vector length to tangent plane
    *
    ********************************************************/
    double denom = sdec * wcs._sdecz + cdec * wcs._cdecz * cradif;
  


    //  Handle vectors too far from axis
    if(denom > tiny)
    {
    }
    else if (denom >= 0.0)
    { 
 
        denom = tiny;
    }
    else if (denom > -tiny) 
    {
      denom = -tiny;
    }
   


    /***************************************************************
    *
    *  Compute tangent plane coordinates (even in dubious cases)
    *
    ***************************************************************/
    x = cdec * sradif / denom;
    y = (sdec * wcs._cdecz - cdec * wcs._sdecz * cradif) / denom;

}
/******************* *************************************
*
* Fortran 95 prototypes
*
**********************************************************/

F90MODPROCPROTO( void, convert_mod, convert1)(double *, double *, double *, double*, double *, double *);
F90MODPROCPROTO( void, convert_mod, convert2)(double *, double*, double*, double*, double *, double *);
F90MODPROCPROTO( void, convert_mod, convert4)(double *, double*, double*, double*, double *, 
                                              double *, double *, double*, double*);
F90MODPROCPROTO( void, convert_mod, convert3)(double *, double *, double *, double*, double *, double *);





/*******************************************************************************************
*
*
*******************************************************************************************/
F90MODPROCPROTO( void, omdistortion_mod, initialisedistortion)(int *, int *, int *, int *);
F90MODPROCPROTO( void, omdistortion_mod, removedistortion)(double *, double *, double *, double *);
F90MODPROCPROTO( void, omdistortion_mod, adddistortion)(double *, double *, double *, double *);


/***********************************************************************************
*
*
***********************************************************************************/
inline void removeDistotion(OMWCS &wcs, double x1, double y1, double x2, double y2)
{

    int x0 = int(wcs.data->crpix1 + 0.000001);
    int y0 = int(wcs.data->crpix2 + 0.000001);
    int binAxis = wcs.data->binAxis;       

    F90MODPROC( omdistortion_mod, initialisedistortion)(&x0, &y0, &binAxis, &binAxis);

    F90MODPROC( omdistortion_mod, removedistortion)(&x1, &y1, &x2, &y2);



}


/***********************************************************************************
*
*
***********************************************************************************/
inline void addDistotion(OMWCS &wcs, double x1, double y1, double x2, double y2)
{

    int x0 = int(wcs.data->crpix1 + 0.000001);
    int y0 = int(wcs.data->crpix2 + 0.000001);
    int binAxis = wcs.data->binAxis;

       

    F90MODPROC( omdistortion_mod, initialisedistortion)(&x0, &y0, &binAxis, &binAxis);

    F90MODPROC( omdistortion_mod, adddistortion)(&x1, &y1, &x2, &y2);



}


/*******************************************************
*
* CompareDates - Compares the first date with the second.
* Returns -1 if the second is less than the first,
* 0 if they are equal and 1 if greater.
*
********************************************************/
int CompareDates(string dateString1, string dateString2)
{    
    int time1[6], time2[6];      // Stores the time from the date string
    char chara;                  // Used for date and time extraction 
    

    sscanf(dateString1.c_str(), "%4d %1c %2d %1c %2d %1c %2d %1c %2d %1c %2d", 
         &time1[0], &chara, &time1[1], &chara, &time1[2], &chara, &time1[3], 
	 &chara, &time1[4], &chara, &time1[5]);



   
    sscanf(dateString2.c_str(), "%4d %1c %2d %1c %2d %1c %2d %1c %2d %1c %2d", 
         &time2[0], &chara, &time2[1], &chara, &time2[2], &chara, &time2[3], 
	 &chara, &time2[4], &chara, &time2[5]);




    for(int i = 0; i < 6; i++)
    {

        int diff = time2[i] - time1[i];
        if (diff != 0) return diff;
 
    }

    return 0; //  Must be equal

}


/*****************************************************************
* Get the start and end time from the stacked files
*
******************************************************************/
void WcsData::setDates(vector<string> &fileList)
{



    DataSet *dataSet = dataSetServer->open( fileList[0], Dal::Read);
      
    startObs = dataSet->attribute("DATE-OBS")->asString();
    endObs = dataSet->attribute("DATE-END")->asString();
    dataSetServer->close(dataSet); // Finished with the file
     

    for(unsigned i = 1; i < fileList.size(); i++)
    {
        DataSet *dataSet = dataSetServer->open( fileList[i], Dal::Read);
      
        string startObs1 = dataSet->attribute("DATE-OBS")->asString();
        string endObs1 = dataSet->attribute("DATE-END")->asString();

	//  cout << i << ") start obs=" << startObs1 << " endobs=" << endObs1 << endl;
        int k = CompareDates(startObs, startObs1);
        
	if (k < 0) startObs = startObs1;

	k = CompareDates(endObs, endObs1);
        
	if (k > 0) endObs = endObs1;
    }
			   
    // cout << "Final start time =" << startObs << " end time=" << endObs << endl;

}

/**************************************************************
* Convert x1, y1 to full frame coords
*
****************************************************************/
void imageToFullFrameCoords(double x, double y, double crpix1, double crpix2, 
                           int binAxis, double &x1, double &y1)
{ 
    x1 = x * double(binAxis) + crpix1;
    y1 = y * double(binAxis) + crpix2;

}



/**************************************************************
* Convert x, y from full frame coords
*
****************************************************************/
inline void fullFrameToImageCoords(double x, double y, double crpix1, double crpix2,
			    int binAxis, double &x1, double &y1)

{
    x1 = (x - crpix1) / double(binAxis);
    y1 = (y - crpix2) / double(binAxis);

}

/******************************************************************
*
* Given x and y converts to RA and DEC in degrees

*
*****************************************************************/   


void convertToCelestial(OMWCS &c, double x, double y, double &ra, double &dec) 
{

    double dx = ( x - c.data->crpix1 ) * c.data->cdelt1;    // Radians    
    double dy = ( y - c.data->crpix2 ) * c.data->cdelt2;    // Radians

    ra = 0; // NB ra and dec will be in degrees
    dec = 0;  
    F90MODPROC( convert_mod, convert1)(&dx, &dy, &c.data->crval1, &c.data->crval2, &ra, &dec);



    //cout << "L292 convertocelestial x=" << x << " y=" << y << endl;
    //cout << "L293 computed ra=" << ra << " dec=" << dec << endl;
    
}

/******************************************************************
*
* Given x and y converts to RA and DEC in radians

*
*****************************************************************/   
inline void convertToCelestial1(OMWCS &c, double x, double y, double &ra, double &dec) 
{

    double dx = ( x - c.data->crpix1 ) * c.data->cdelt1;    // Radians    
    double dy = ( y - c.data->crpix2 ) * c.data->cdelt2;    // Radians

    ra = 0;  // NB Ra and dec will be in radians
    dec = 0;  
    F90MODPROC( convert_mod, convert3)(&dx, &dy, &c.data->crval1, &c.data->crval2, &ra, &dec);

    //cout << "L313 convertocelestial x=" << x << " y=" << y << endl;
    //cout << "L314 computed ra=" << ra << " dec=" << dec << endl;
    
}


/******************************************************************
*
* Given x and y converts to RA and DEC in degrees
*
*****************************************************************/   
inline void convertToCelestial(WcsData1 &c, double x, double y, double &ra, double &dec) 
{

    double dx = ( x - c.crpix1 ) * c.cdelt1;    // Radains    
    double dy = ( y - c.crpix2 ) * c.cdelt2;    // Radians

    ra = 0; // NB RA and Dec will be in degrees
    dec = 0;  
    F90MODPROC( convert_mod, convert1)(&dx, &dy, &c.crval1, &c.crval2, &ra, &dec);

    //cout << "L334 omwcscc convertocelestial x=" << x << " y=" << y << endl;
    //cout << "L335 computed ra=" << ra << " dec=" << dec << endl;
    // cout << "cdelt1=" << c.cdelt1 << " cdelt2=" << c.cdelt2 << endl;     
    // cout << "crpix1=" << c.crpix1 << " crpix2=" << c.crpix2 << endl;
}

/*****************************************************************
*
* Converts from RA, DEC to x and y
*
*
******************************************************************/
inline void convertFromCelestial(double ra, double dec, OMWCS& c, double &x, double &y)
{
            
    /**************************************************************
    *
    *
    *
    ****************************************************************/
    F90MODPROC( convert_mod, convert4)(&ra, &dec, &c.data->crval1, 
                                       &c.data->crval2, 
                                       &c.data->crpix1, 
                                       &c.data->crpix2, 
                                       &c.data->cdelt1, &x, &y);

    //cout << "L360 Wcs.cc- convert from celestial - ra=" << ra << " dec=" << dec << " x=" << x << " y=" << y << endl;

}

/********************************************************
*
*
*
*
*******************************************************/
OMWCS::OMWCS(int k) 
{
  //   cout << "Line 284 WCS Constructor  k=" << k << endl; 

    imageType = k;
    if (k == 0)
    {
        data = new WcsData1;
    }
    else
    {
      //    cout << "Constructing wcsdara2\n";
        data = new WcsData2;
    } 
};    

/*********************************************************************
*
*
*********************************************************************/

OMWCS::OMWCS() { data = 0; };


OMWCS::~OMWCS() { if (data) { delete data; data = 0;}; };
  
/*******************************************************
*
*
*
*
*******************************************************/

OMWCS::OMWCS(const OMWCS &c)
{

  //cout << "Line 316 WCS copy constructor\n";

    if(this == &c) 
    {  
      //      cout << "returning at line 320\n";
        return;
    }
    //   cout << "Line 422\n";
    //   cout << "Line 319 WCS copy constructor imagetype=" << c.imageType << " data=" << data << endl;

    //  cout << "Line 425\n";



  // cout << data << endl;
  //   if(data) delete data;

    //cout << "Line 320\n";

    imageType = c.imageType;
    if (imageType == 0)
    {
      //cout << "Line 428 WCS copy constructor\n";

        data = new WcsData1;
    }
    else
    {
      // cout << "Line 434 WCS copy constructor\n";


        data = new WcsData2;
    } 


    //  cout << "Line 441 WCS copy constructor\n";

    
    //cout << "L444 c data xLeft=" << c.data->xLeft << " xRight="<< c.data->xRight << endl;

    data->startObs = c.data->startObs;
    
    data->endObs = c.data->endObs;
    data->_obsId = c.data->_obsId;
    data->object = c.data->object;

    data->nx = c.data->nx; 
    data->ny = c.data->ny; 
    data->cdelt1 = c.data->cdelt1; 
    data->cdelt2 = c.data->cdelt2; 

    data->crpix1 = c.data->crpix1; 
    data->crpix2 = c.data->crpix2; 

    data->crval1 = c.data->crval1; 
    data->crval2 = c.data->crval2; 
    data->_cdecz = c.data->_cdecz; 
    data->_sdecz = c.data->_sdecz; 
    data->_raPnt = c.data->_raPnt; 
    data->_decPnt = c.data->_decPnt; 
    data->_paPnt = c.data->_paPnt; 
    data->_mjdMid = c.data->_mjdMid; 
    data->_astrometry = c.data->_astrometry;



    // cout << "Line 474 WCS copy constructor\n";
 
    data->binAxis = c.data->binAxis; 
    

    data->xLeft= c.data->xLeft;
    data->xRight = c.data->xRight;
    data->yBottom = c.data->yBottom;
    data->yTop = c.data->yTop;
    data->_area = c.data->_area;

   
    data->_dx = c.data->_dx;
    data->_dy = c.data->_dy;

    //cout << "L487 c.data->frameTime=" << c.data->frameTime << " exposure=" << c.data->exposure << endl;
    data->deadFrac = c.data->deadFrac;
    data->frameTime = c.data->frameTime;
    data->exposure = c.data->exposure;


    data->filter = c.data->filter;
    
   
   
};

/*******************************************************
*
*
*
*
*******************************************************/

OMWCS& OMWCS::operator=(const OMWCS &c)
{

 

    if(this == &c) 
    {  
        return *this;
    }
    
    if(data) delete data;


    imageType = c.imageType;
    if (imageType == 0)
    {
        data = new WcsData1;
      

    }
    else
    {
   
        data = new WcsData2;
    } 

 
    data->startObs = c.data->startObs;
   
    data->endObs = c.data->endObs;
    data->_obsId = c.data->_obsId;
    data->object = c.data->object;


    data->nx = c.data->nx; 
    data->ny = c.data->ny; 
    data->cdelt1 = c.data->cdelt1; 
    data->cdelt2 = c.data->cdelt2; 

    data->crpix1 = c.data->crpix1; 
    data->crpix2 = c.data->crpix2; 

    data->crval1 = c.data->crval1; 
    data->crval2 = c.data->crval2; 

    data->_cdecz = c.data->_cdecz; 
    data->_sdecz = c.data->_sdecz; 

    data->_raPnt = c.data->_raPnt; 
    data->_decPnt = c.data->_decPnt; 
    data->_paPnt = c.data->_paPnt; 
    data->_mjdMid = c.data->_mjdMid; 

    data->_astrometry = c.data->_astrometry;


    data->binAxis = c.data->binAxis; 


    data->xLeft = c.data->xLeft;
    data->xRight = c.data->xRight;
    data->yBottom = c.data->yBottom;
    data->yTop = c.data->yTop;

    data->_area = c.data->_area;


    data->_dx = c.data->_dx;
    data->_dy = c.data->_dy;

    data->deadFrac = c.data->deadFrac;
    data->frameTime = c.data->frameTime;
    data->exposure = c.data->exposure;


    data->filter = c.data->filter;
    
    return *this;
};








static double meanCrval1 = 0.0, meanCrval2 = 0.0;
static double meanCrpix1 = 0.0, meanCrpix2 = 0.0;

static int count1 = 0;


/****************************************************
*
*
****************************************************/

WcsData1& WcsData1::operator=(const WcsData1& rhs)
{


    if (this == &rhs) return *this;

    startObs = rhs.startObs;
    endObs = rhs.endObs;
    _obsId = rhs._obsId;

    object = rhs.object;


    nx = rhs.nx; 
    ny = rhs.ny; 
    cdelt1 = rhs.cdelt1; 
    cdelt2 = rhs.cdelt2; 

    crpix1 = rhs.crpix1; 
    crpix2 = rhs.crpix2; 

    crval1 = rhs.crval1; 
    crval2 = rhs.crval2; 

 

    _cdecz = rhs._cdecz;
    _sdecz = rhs._sdecz;

    _raPnt = rhs._raPnt; 
    _decPnt = rhs._decPnt; 
    _paPnt = rhs._paPnt; 

    _mjdMid = rhs._mjdMid; 

    _astrometry = rhs._astrometry;

    binAxis = rhs.binAxis; 


    xLeft = rhs.xLeft;
    xRight = rhs.xRight;
    yBottom = rhs.yBottom;
    yTop = rhs.yTop;
    _area = rhs._area;


    _dx = rhs._dx;
    _dy = rhs._dy;


    deadFrac = rhs.deadFrac;
    frameTime = rhs.frameTime;
    exposure = rhs.exposure;


    filter = rhs.filter;
    
    return *this;

};




/**********************************************************************
*
*
*
*
**********************************************************************/
bool WcsData1::load( string file)
{



  //    cout << "omwcs.cc loading data from file " << file << endl;


    DataSet *dataSet = dataSetServer->open( file, Dal::Read);


    if(dataSet->hasAttribute("RA_PNT"))       
        _raPnt = dataSet->attribute("RA_PNT")->asReal();
    if(dataSet->hasAttribute("DEC_PNT"))     
        _decPnt = dataSet->attribute("DEC_PNT")->asReal();

    if(dataSet->hasAttribute("PA"))     
        _paPnt = dataSet->attribute("PA")->asReal();

  

    if(dataSet->hasAttribute("POSCOROK"))  
    {   

        /********************************************************** 
	*
        * NB RAOFFSET is the computed astrometry x offset (arcsecs) 
        * at the OM pointing direction
        ***********************************************************/

    

        _astrometry._poscorrOkay = dataSet->attribute("POSCOROK")->asBool();


        _astrometry._raOffset = dataSet->attribute("RAOFFSET")->asReal() / cos(_decPnt * degreesToRadians);


	//   _astrometry._raOffsetError = dataSet->attribute("RAOFFERR")->asReal();

        _astrometry._decOffset = dataSet->attribute("DEOFFSET")->asReal();



	// _astrometry._decOffsetError = dataSet->attribute("DEOFFERR")->asReal();
	// _astrometry._rmsResidual = dataSet->attribute("RMSRESID")->asReal();
	//  _astrometry._nMatches = dataSet->attribute("NMATCHES")->asInt();

    }    

    if ( dataSet->hasAttribute("CDELT1") )
    {
   
        cdelt1 = dataSet->attribute("CDELT1")->asReal() * degreesToRadians;  // RA plate-scale (degrees and negative)
        cdelt2 = dataSet->attribute("CDELT2")->asReal() * degreesToRadians;  // Dec plate-scale (degrees)
        crpix1 = dataSet->attribute("CRPIX1")->asReal();
        crpix2 = dataSet->attribute("CRPIX2")->asReal();
        crval1 = dataSet->attribute("CRVAL1")->asReal() * degreesToRadians;
        crval2 = dataSet->attribute("CRVAL2")->asReal() * degreesToRadians;
       
     
    }
    else
    {
        const Block *block = dataSet->block(0);
        if ( !block->hasAttribute( "CDELT1" ) )
	{


	      errstr << "No attribute CDELT1 in file  " << file << "!- terminating program\n" 
	     << Sas::message(Msg::App, Msg::Sparse);

  

       

             dataSetServer->close(dataSet); // Finished with the file
             Sas::fatal("Bad file", "Attribute CDELT1 does not exist in file");


        }

        cdelt1 = block->attribute("CDELT1")->asReal() * degreesToRadians;  // RA plate-scale (degrees and negative)
        cdelt2 = block->attribute("CDELT2")->asReal() * degreesToRadians;  // Dec plate-scale (degrees)
        crpix1 = block->attribute("CRPIX1")->asReal();
        crpix2 = block->attribute("CRPIX2")->asReal();
        crval1 = block->attribute("CRVAL1")->asReal() * degreesToRadians;

        crval2 = block->attribute("CRVAL2")->asReal() * degreesToRadians;




        if(block->hasAttribute("RA_PNT"))       
            _raPnt = block->attribute("RA_PNT")->asReal();
        if(block->hasAttribute("DEC_PNT"))     
            _decPnt = block->attribute("DEC_PNT")->asReal();
        if(block->hasAttribute("PA"))     
            _paPnt = block->attribute("PA")->asReal();

        if(block->hasAttribute("POSCOROK"))  
        {   

	 
            _astrometry._poscorrOkay = block->attribute("POSCOROK")->asBool();
	   

	    /* 2011-02-21 vny avoiding a crash if there is no RAOFFSET*/
	    if(block->hasAttribute("RAOFFSET")){
	      _astrometry._raOffset = block->attribute("RAOFFSET")->asReal() / cos(_decPnt * degreesToRadians);
	    
	    } else {
	      if(block->hasAttribute("RA_OFF")){
	
		_astrometry._raOffset = block->attribute("RA_OFF")->asReal() / cos(_decPnt * degreesToRadians);
	      } else {
	
		_astrometry._raOffset=0.0;
	      }
	    }

	    if(block->hasAttribute("DEOFFSET")){
	     
	      _astrometry._decOffset = block->attribute("DEOFFSET")->asReal();
	    } else {
	     
	      if(block->hasAttribute("DEC_OFF")){
	
		_astrometry._decOffset = block->attribute("DEC_OFF")->asReal();
	      } else {
	
		_astrometry._decOffset = 0.0;
	      }
	    }



	    // _astrometry._decOffsetError = block->attribute("DEOFFERR")->asReal();
	    // _astrometry._rmsResidual = block->attribute("RMSRESID")->asReal();
	    // _astrometry._nMatches = block->attribute("NMATCHES")->asInt();



       }    

    }


  
    /*********************************************************************
    *
    * Get the binning factor
    *
    *********************************************************************/

    binAxis = (int)pow(2, (double)dataSet->attribute("BINAX1")->asInt());
    binBpe = dataSet->attribute("BINBPE")->asBool();
    if( binBpe)
    { 
        binAxis *= 2; 
     
    }

    _cdecz = cos(crval2);
    _sdecz = sin(crval2);

    
    ++count1;

    meanCrval1 += crval1;
    meanCrval2 += crval2;

    meanCrpix1 += crpix1;
    meanCrpix2 += crpix2;


    /*****************************************************************
    *
    * Get the size of the image
    *
    *****************************************************************/
    const Array* array = dataSet->array("PRIMARY");
  
    longVec size = array->size();
  

    nx = (int)size[1];
    ny = (int)size[0];

    //cout << "L864 " << endl;
    if( dataSet->hasAttribute( "DEADFRAC" ) )
    {
        deadFrac = dataSet->attribute("DEADFRAC")->asReal();
        frameTime = dataSet->attribute("FRAMTIME")->asReal();

        // SPR-7299 : set deadfrac from the frame time
        if (frameTime>0.0)
           deadFrac = 0.174 / frameTime;
        else
           deadFrac = 0.0;

	//cout << "L869b deadFrac=" << deadFrac << " frameTime=" << frameTime << endl;
    }
    else
    {
      //cout << "L873 no attribute DEADFRAC " << endl;
        deadFrac = 0.0;
        frameTime = 0.0; 

    }
    exposure = dataSet->attribute("EXPOSURE")->asReal();
 

    filter = dataSet->attribute("FILTER")->asString();
   
    if (dataSet->hasAttribute("OBJECT"))
    {
        object = dataSet->attribute("OBJECT")->asString();
    }

    string startObs = dataSet->attribute("DATE-OBS")->asString();
    string endObs = dataSet->attribute("DATE-END")->asString();
    _obsId = dataSet->attribute("OBS_ID")->asString();


    dataSetServer->close(dataSet); // Finished with the file
   


    /****************************************************************************
    *
    * xLeft is the RA of the left-most image pixel, etc
    *
    *****************************************************************************/
   
    //  cout << "omwcss.cc line 910 converting xleft, xright, ybottom and ytop\n";

    /**********************************************************
    *
    * Compute the RA/DEC coordinates of the image edges
    *
    ************************************************************/
    getCornerCoords();





    //    cout << "omwcss line 952 after getcornercoords\n";
    // cout << "omwcs.cc Line 918  xLeft=" << xLeft * radiansToDegrees  << " yTop=" << yTop * radiansToDegrees  << endl;
    // cout << "omwcs.cc Line 918 xright=" << xRight * radiansToDegrees  << " yBottom=" << yBottom * radiansToDegrees << endl;
    // cout << "crval1=" << crval1 * radiansToDegrees << " crval2=" << crval2 * radiansToDegrees << endl;


    if(xRight > xLeft) xRight -= 360.0;

    _area = (xLeft - xRight) * (yTop - yBottom);

    /*******************************************************************
    *
    * Need the mid-time of the observation
    *
    *
    *******************************************************************/

    double mjdStart = ModifiedJulianDate( startObs );
    double mjdEnd =   ModifiedJulianDate( endObs );

    _mjdMid = (mjdStart + mjdEnd) / 2.0;


  
    return true;

}

/***************************************************************************
*
*
*
*
*
***************************************************************************/
void WcsData1::setImageParameters()
{
  
  //cout << "***************************************************\n";
  //cout << "L962 wcs.cc wcsdata1 setimageparameters \n";
  
  //   return;
  //cout << "L968 original crval1=" << crval1 << " crval2=" << crval2 << endl;
  //cout << "L969 original crpix11=" << crpix1 << " crpix2=" << crpix2 << endl;

  //*************************************************************
  // NB should the cos be in the formulaa below ????
  //*************************************************************
  
  //cout << "L978 xleft=" << xLeft * radiansToDegrees  << " right=" << xRight * radiansToDegrees << endl;
  //cout << "L979 ybottom=" << yBottom * radiansToDegrees << " top=" << yTop * radiansToDegrees << endl;
  //cout << "L980 cdelt1=" << cdelt1 << " cdelt2=" << cdelt2 << endl;
  
  if ((xLeft * radiansToDegrees) > 360.0){
    xLeft-=360.0 *degreesToRadians;
  }
  if ((xRight * radiansToDegrees) > 360.0){
    xRight-=360.0*degreesToRadians;
  }

  //    crval1 = (xLeft + xRight) / 2.0;  
  //   if(crval1 > TWO_PI) crval1 -= TWO_PI;
  //  crval2 = (yBottom + yTop) / 2.0;

  double c = min( abs(yBottom), abs(yTop) );
  //cout << "L989 c=" << c << " yBottom=" << yBottom << " yTop=" << yTop << endl;
  /* There is a possibility of crossing the line of 24 hours. 
     This should be taken into account. NB: the x-coordinate is 
     increasing in the reverse direction */
  //cout << "L994 xLeft=" << xLeft << " xRight=" << xRight << endl;
  if (xLeft < xRight){
    xLeft+=360.0 *degreesToRadians;
  }
  //cout << "L994 xLeft=" << xLeft << " xRight=" << xRight << endl;
  int nxInitial=(int)ceil( (xLeft - xRight) * abs(cos(c))  / -cdelt1);
  int nyInitial=(int)ceil( (yTop - yBottom) / cdelt2); 
  //cout << "L997 nxInitial=" << nxInitial << " nyInitial=" << nyInitial << endl;
  // RDS remove the 4096 limit
/*
  if (nxInitial>4096){
    //cout << "L999 setting nx to 4096 \n";
      nxInitial=4096;
      xRight=xLeft+(float)nxInitial*cdelt1/abs(cos(c));
    }
    if (nyInitial>4096){
      nyInitial=4096;
      yTop=yBottom+(float)nyInitial*cdelt2;
    }
  */
  // RDS remove the 4096 limit
    
    //cout << "L1002 nxInitial=" << nxInitial << " nyInitial=" << nyInitial << endl;
    //cout << "L1003 xRight=" << xRight* radiansToDegrees << " yTop=" << yTop* radiansToDegrees << endl;
    
    crval1 = xLeft;
    crval2 = yBottom;
    crpix1 = 0.5;
    crpix2 = 0.5;

    nx = (int)ceil( (xLeft - xRight) * abs(cos(c))  / -cdelt1);

    //nx+=10;
 
    ny = (int)ceil( (yTop - yBottom) / cdelt2); 

    //ny+=10;
    //cout << "L1017 initial nx=" << nx << " ny=" << ny << endl;
    //cout << "L1018 cdelt1=" << cdelt1 << " cdelt2=" << cdelt2 << endl; 

    // NB HAVE ADDED THE FOLLOWING
    //    double dx1 = 0.0, dy1= 0.0;
    // celestialToTangentPlane( *this, crval1, crval2, dx1, dy1);


  
    //cout << "L1035 wcs.cc setImageParameters\n";
    
    //cout << "xleft=" << xLeft * radiansToDegrees  << " right=" << xRight * radiansToDegrees << endl;
    //cout << "ybottom=" << yBottom * radiansToDegrees << " top=" << yTop * radiansToDegrees << endl;
    //cout << "cdelt1=" << cdelt1 << " cdelt2=" << cdelt2 << endl;
    //cout << "crval1=" << setprecision(10) << crval1 * radiansToDegrees << " crval2=" 
     //<< setprecision(10) << crval2 * radiansToDegrees  << endl;
    //cout << "crpix1=" << crpix1 << " crpix2=" << crpix2 << endl;
    //cout << "nx=" << nx << " ny=" << ny << endl;
    

    //   crval1 = xLeft;
    //  crval2 = yBottom;
    //  crpix1 = 0.5;
    // crpix2 = 0.5;


    _cdecz = cos(crval2); 
    _sdecz = sin(crval2); 

}

/*******************************************************************
*
*
* For sky images
* Computes the RA and Dec coordinates of each corner of the image
*
*
******************************************************************/
void WcsData1::getCornerCoords()
{
  
  // cout << "Wcsdata1 getcornercoords nx=" << nx << " ny=" << ny << endl;
  //  cout << "binaxis=" << binAxis << endl;
  //  cout << "cdelt1=" << cdelt1 << " cdelt2=" << cdelt2 << endl; 
 
 //cout << "crpix1=" << crpix1 << " crpix2=" << crpix2 << endl;


    /**************************************************************
    *
    * Get left-hand bottom corner RA and DEC of the image 
    *
    **************************************************************/
    double x = 0.5, y = 0.5;
    convertToCelestial(*this, x, y, xLeft, yBottom);

  
    /**************************************************************
    *
    * Get right-hand bottom corner RA and DEC of the image 
    *
    **************************************************************/
    x = double(nx) + 0.5;
    convertToCelestial(*this, x, y, xRight, yBottom);
 

    /**************************************************************
    *
    * Get left-hand top corner RA and DEC of the image 
    *
    **************************************************************/

    x = 0.5;
    y = double(ny) + 0.5;
    
    convertToCelestial(*this, x, y, xLeft, yTop);

    /**************************************************************
    *
    * Get right-hand top corner RA and DEC of the image 
    *
    **************************************************************/
    x = double(nx) + 0.5;
    y = double(ny) + 0.5;
   
    convertToCelestial(*this, x, y, xRight, yTop);
        
    //    cout << "nx=" << nx << " ny=" << ny << endl; 

    if(xRight > xLeft) xRight -= 360.0;

    xLeft   *= degreesToRadians;
    xRight  *= degreesToRadians;
    yBottom *= degreesToRadians;
    yTop    *= degreesToRadians;


    //  cout << "Image- xleft (radians)=" << setprecision(14) << xLeft << endl;      
    // cout << "Image- xright=" << setprecision(14) << xRight << endl;      
    // cout << "Image- ytop=" << setprecision(14) << yTop << endl;      
    // cout << "Image- ybottom=" << setprecision(14) << yBottom << endl;      


}
/***************************************************************
*
*
*
*
*
***************************************************************/
void WcsData1::convert() {};

/**************************************************************************
*
* For sky-images
*
* Convert image 1 coordinates  x1, y1 on image 1 to image 2 coordinates x2, y2  
* 
* 
*
***************************************************************************/
inline void WcsData1::convertX1Y1ToX2Y2(double x1, double y1, 
                OMWCS &c2, double &x2, double &y2)
{

    /****************************************************************
    *
    * First convert x and y to ra and dec
    *
    *****************************************************************/
    double ra = 0.0, dec = 0.0;


    double dx1 = ( x1 - crpix1 ) * cdelt1;    // Radians    
    double dy1 = ( y1 - crpix2 ) * cdelt2;    // Radians



    tangentPlaneToCelestial(*this, dx1, dy1, ra, dec);

   
    /***************************************************
    *
    *
    * Convert Ra and dec to x2 and y2
    *
    ***************************************************/
 
    x2 = 0.0, y2 = 0.0;

 

    WcsData1 *wcs = static_cast<WcsData1*>(c2.data);
  
    celestialToTangentPlane( *wcs, ra, dec, dx1, dy1);


    x2 = c2.data->crpix1 + dx1 / c2.data->cdelt1;  
    y2 = c2.data->crpix2 - dy1 / c2.data->cdelt1;


}
/**********************************************************************
*
*
*
*
**********************************************************************/

bool WcsData1::withinImage(double x, double y)
{

    if ( x >= 0.5 && y >= 0.5 && x <= double(nx) + 0.5 &&  y <= double(ny) + 0.5)
        return true;
    else
        return false;


}

/**********************************************************************************
*
* Creates the mosaiced sky image
*
*
**********************************************************************************/
// 2014-02-02 adding max deadFrac and frameTime
void WcsData1::createOutputFile(string fileName,
                                       vector< vector<real32> > &image, 
                                       vector< vector<real32> > &exposureImage,
                                       vector< vector<int16> > &qualityImage,
				//vector<string> &fileList)
				vector<string> &fileList, double deadFracMax, double frameTimeMax)
{

    if(fileList.size() == 0)
    {
        errstr << "L1217 Cannot create output image- no image files"
	       << Sas::warning("nofiles");

        return;

    }
    int numStacked = (int)fileList.size();
    //cout << "L1222 numStacked=" << numStacked << endl;
    //cout << "L1223 Creating file " << fileName << endl;
    //cout << "L1226 numStacked=" << numStacked << " frameTimeMax=" << frameTimeMax << endl;
    DataSet * set  = dataSetServer->open(fileName, Dal::Create);
  
    int nx = (int)image.size();
    int ny = (int)image[0].size();
  

    /*********************************************************
    *
    * Divide the image by the exposure image
    *
    *********************************************************/
    normalizeImage(image, exposureImage);


    addBlock(set, "PRIMARY", image, 0);
    addBlock(set, "EXPOSURE", exposureImage, 1);

    addBlock(set, "QUALITY", qualityImage, 2);


    // ****************************
    // Copy attributes from source.
    // ****************************
    //    vector<string> imageList = stringListParameter( "imagesets" );

    //  DataSet * set1 = dataSetServer->open(imageList[0], Dal::Read);
    // for( unsigned int i = 0; i < set1->attributes(); ++i )
    //  {
    //	set->addAttribute( set1->attribute( i ) );
    //    cout << i << ") Attribute=" << set1->attribute(i)->name() << endl;

    //    } 



    /*******************************************************************
    *
    *
    *******************************************************************/
    *set->addAttribute("INSTRUME", "") = "OM";

    *set->addAttribute("DATAMODE", "") = "IMAGING";

    *set->addAttribute("FILTER", "") = filter;

    /****************************************
    *
    * Add various keywords to set.
    ****************************************/
      //    call setAttribute(set, "RADECSYS", "FK5", " ")
      //   call setAttribute(set, "EQUINOX", 2000, "", "Equinox of celestial coordinate system")
      // call setAttribute(set, "EXPOSURE", obsDetails%exposure, "secs", "Exposure")
      // call setAttribute(set, "DATE-OBS", obsDetails%startObs, "Start date of the first exposure")
      //call setAttribute(set, "DATE-END", obsDetails%EndObs,   "End date of the last exposure")
      // call setAttribute(set, "OBSERVER", obsDetails%observer, "Name of PI")
      //   call setAttribute(set, "OBJECT",   obsDetails%object, "Name of object")


    /*****************************************************************
     Added by CB on July 18 2008
    
    ******************************************************************/
      //  call setAttribute(set, "DATAMODE",   "MOSAICED", "OM imagine mode")
      // call setAttribute(set, "DEADFRAC",   obsDetails%DeadFrac, "OM imagine mode")
      // call setAttribute(set, "FRAMTIME",   obsDetails%FrameTime, "OM imagine mode")
      //  call setAttribute(set, "WINDOWX0",  0, "OM imagine mode")






 //   call setAttribute(set, "WINDOWDX",  image%dims(0), "OM imagine mode")
 //   call setAttribute(set, "WINDOWY0",  0, "OM imagine mode")
 //   call setAttribute(set, "WINDOWDY",  image%dims(1), "OM imagine mode")
 //   call setAttribute(set, "BINAX1",  obsDetails%binax1, "X binning factor")
 //   call setAttribute(set, "BINAX2",  obsDetails%binax1, "Y binning factor")
 //   call setAttribute(set, "BINBPE",  "FALSE", "Binning")



 //   call setAttribute(set, "RA_PNT",  image%ra_pnt, "Mean RA pointing direction")
 //   call setAttribute(set, "DEC_PNT",  image%dec_pnt, "Mean Dec pointing direction")
 //   call setAttribute(set, "PA",  image%pa, "Mean position angle of pointing direction")



  //  call setAttribute(set, "INSTRUME", obsDetails%instrument, "Instrument")
  //  call setAttribute(set, "FILTER", obsDetails%filter, "Filter")

 


    *set->addAttribute("OBS_ID", "") = _obsId;

  
    *set->addAttribute("OBJECT", "") = object;
 
    *set->addAttribute("DATE-OBS", "") = startObs;
    *set->addAttribute("DATE-END", "") = endObs;


    
    binAxis /= 2;
    

    *set->addAttribute("BINAX1", "") = binAxis;
    *set->addAttribute("BINAX2", "") = binAxis;
    *set->addAttribute("BINBPE", "") = false;

    *set->addAttribute("EXPOSURE", "") = 1;

    //cout << "L1336 deadFrac=" << deadFrac << " frameTime=" << frameTime <<endl;

    //*set->addAttribute("DEADFRAC", "") = deadFrac;
    // *set->addAttribute("FRAMTIME", "") = frameTime;
    *set->addAttribute("DEADFRAC", "") = deadFracMax;
    *set->addAttribute("FRAMTIME", "") = frameTimeMax;
    

    *set->addAttribute("WINDOWX0", "") = 0;
    *set->addAttribute("WINDOWDX", "") = nx;
    *set->addAttribute("WINDOWY0", "") = 0;
    *set->addAttribute("WINDOWDY", "") = ny;

 


    /*************************************************************************
    *
    *
    *
    **************************************************************************/

    *set->addAttribute("EQUINOX", "Equinox of celestial coordinate system") = 2000;
 
 
    *set->addAttribute( "CDELT1", "Plate-scale in x direction") = cdelt1 * radiansToDegrees;
    *set->addAttribute( "CDELT2", "Plate-scale in y direction") = cdelt2 * radiansToDegrees;
    *set->addAttribute( "CTYPE1", "") = "RA---TAN";
    *set->addAttribute( "CTYPE2", "") = "DEC--TAN";


    *set->addAttribute( "CRPIX1", "Reference x pixel") = crpix1;
    *set->addAttribute( "CRPIX2", "Reference y pixel") = crpix2;

    *set->addAttribute( "CRVAL1", "RA at reference  x pixel") = crval1 * radiansToDegrees;
    *set->addAttribute( "CRVAL2", "DEC at reference  y pixel") = crval2 * radiansToDegrees;




    *set->addAttribute( "PA", "Mean RA pointing direction") =  _paPnt;
 
    *set->addAttribute( "RA_PNT", "Mean RA pointing direction") =  crval1 * radiansToDegrees;
    *set->addAttribute( "DEC_PNT", "Mean Dec pointingDirection") = crval2 * radiansToDegrees;




    if (_astrometry._poscorrOkay)
    {
        *set->addAttribute( "POSCOROK", "Images were astrometry corrected") = _astrometry._poscorrOkay;

    }
    else
    {
        *set->addAttribute( "POSCOROK", "Images were not astrometry corrected") = _astrometry._poscorrOkay;

    }
 
    *set->addAttribute( "RAOFFSET", "Mean RA offset applied to images") = _astrometry._raOffset;

    *set->addAttribute( "DEOFFSET", "Mean DEC offset applied to images") = _astrometry._decOffset;

    *set->addAttribute( "NSTACKED", "Number of images stacked") = numStacked;

    //  for(unsigned i = 0; i < fileList.size(); i++)
    // {
    //  char text[100];
    //  sprintf(text, "IMAGE%03d", i + 1);
    //  *set->addAttribute( text, "Image stacked") = fileList[i];
    //  }

    dataSetServer->close(set); // Finished with the file
    


}


/*************************************************************
*
* This is for sky images
*
*
**************************************************************/
void WcsData1::convertToLowResImage(vector< vector<float> > &image, 
                vector< vector<float> > &exposureImage,
		vector< vector<int16> > &qualityImage)

{

    if(binAxis == 2) return;

     nx = 0, ny = 0;
    if ( (image.size() / 2) * 2 == image.size() )
      nx = image.size() / 2;
    else
      nx = image.size() / 2 + 1;


    if ( (image[0].size() / 2) * 2 == image[0].size() )
      ny = image[0].size() / 2;
    else
      ny = image[0].size() / 2 + 1;

    vector< vector<float> > tempImage(nx, vector<float>(ny, -9999.9));
    vector< vector<float> > tempExposureImage(nx, vector<float>(ny, -9999.9));

    vector< vector<int16> > tempQualityImage(nx, vector<int16>(ny, 0));


    int nx1 = image.size() - 1;
    int ny1 = image[0].size() - 1;


    for (int i = 0; i <= nx1; i+= 2)
    {

        int i1 = i / 2;

        for (int j = 0; j <= ny1; j+= 2)
        {
  
 
            int j1 = j / 2;
            int count = 0;
            double sum1 = 0.0, sum2 = 0.0;

            int16 qFlag;
            for(int k = i; k <= min(i + 1, nx1); k++)
	    {
                for(int l = j; l <= min(j + 1, ny1); l++)
		{
                  if (!isNull(image[k][l] ) )
		  {
		      sum1 += (double)image[k][l];
                      sum2 += (double)exposureImage[k][l];
                      ++count;
                
                      for(int n = 0; n < 12; n++)
		      {
		          if (bitSet( qualityImage[k][l], n ) )
		          {
			    setBit(qFlag, n);
                          }
		      }

		  }


		} 

            }  

            if(count > 0)
	    {
	      tempImage[i1][j1] = sum1;
              tempExposureImage[i1][j1] = sum2 / (double)count;
              tempQualityImage[i1][j1] = qFlag; 
	    } 
	}
    }

    image = tempImage;
    exposureImage = tempExposureImage;
    qualityImage = tempQualityImage;


    /************************************************************
    *
    *
    *
    *
    **************************************************************/
    cdelt1 *= 2.0;
    cdelt2 *= 2.0;
    crpix1 /= 2.0;
    crpix2 /= 2.0;
    binAxis = 2;
  

}

/*************************************************************
*
* For sky images
*
***************************************************************/
void WcsData1::correctReferencePixel(double dx1, double dy1)
{
 



    crpix1 -= dx1;
    crpix2 -= dy1;


  



}

/*************************************************************
*
* For sky images
* NB _dx and _dy are in unbinned pixels
*
***************************************************************/
void WcsData1::correctReferencePixel()
{

    if(binAxis == 1)
    {   
        crpix1 += _dx;
        crpix2 += _dy;
    }
    else
    {   
        crpix1 += _dx / 2.0;
        crpix2 += _dy / 2.0;
    }

}

/*******************************************************************
*
* For sky images
*
*
********************************************************************/
void WcsData1::updateReferencePixelCorrection(OMWCS &c, double dx, double dy)
{


    /***************************************************************
    *
    * All offsets (dx and dy) will be in unbinned pixels
    *
    ****************************************************************/
    if(c.data->binAxis  == 1)
    {
        if(binAxis  == 1)
        {
            _dx += c.data->_dx + dx;
            _dy += c.data->_dy + dy;
        } 
        else
        {
            _dx += c.data->_dx + dx * 2.0;
            _dy += c.data->_dy + dy * 2.0;
        } 
    }
    else
    {

        if(binAxis  == 1)
        {
            _dx += c.data->_dx + dx;
            _dy += c.data->_dy + dy;
        } 
        else
        {
            _dx += c.data->_dx + dx * 2.0;
            _dy += c.data->_dy + dy * 2.0;
        } 

      
    } 
 

}


/*******************************************************************************
*
*
*
*
********************************************************************************/
double WcsData1::distance(double x, double y, OMWCS &c, double &x2, double &y2)
{

    double x1 = 0.0, y1 = 0.0; 
    convertX1Y1ToX2Y2(x, y, c, x1,  y1);

    double dx = x1 - x2;
    double dy = y1 - y2;
    return hypot(dx, dy);

    


}

/*******************************************************************************
*
*
*
*
********************************************************************************/
double WcsData1::overlapFraction(OMWCS &c2)
{
   
    double x1 = min(xLeft,  c2.data->xLeft);
    double x2 = max(xRight,  c2.data->xRight);

    double y1 = max(yBottom,  c2.data->yBottom);
    double y2 = min(yTop,  c2.data->yTop);


    //  cout << "Line 1340 overlapfraction x1=" << x1 << " x2=" << x2 << " y1=" << y1 << " y2=" << y2 << endl;
    if( (x1 - x2) < cdelt2 || (y2 - y1) < cdelt2) return 0.0;


    /*************************************************
    * Compute the overlapping image area
    *
    **************************************************/
    double area1 = (x1 - x2) * (y2 - y1);
    

    /*************************************************
    *
    * Compute the area of this image
    *
    **************************************************/
    double area2 = ((xLeft - xRight) * (yTop - yBottom)); 


    /*************************************************
    *
    * Compute the area of the other image (c2)
    *
    **************************************************/
    double area3 = ((c2.data->xLeft - c2.data->xRight) * (c2.data->yTop - c2.data->yBottom));


    //   cout << "area1=" << area1 << " area2=" << area2 << " area3=" << area3 << " fraction=" <<   area1 / max(area2, area3) << endl; 
    return area1 / max(area2, area3); 
  
}

/****************************************************
*
*
*
*
****************************************************/
WcsData2& WcsData2::operator=(const WcsData2& rhs)
{


    if (this == &rhs) return *this;

    //   cout << "wcsdata1 operator=\n";
    startObs = rhs.startObs;
    endObs = rhs.endObs;
    _obsId = rhs._obsId;



    nx = rhs.nx; 
    ny = rhs.ny; 
    cdelt1 = rhs.cdelt1; 
    cdelt2 = rhs.cdelt2; 

    crpix1 = rhs.crpix1; 
    crpix2 = rhs.crpix2; 

    crval1 = rhs.crval1; 
    crval2 = rhs.crval2; 

    binAxis = rhs.binAxis; 


    xLeft = rhs.xLeft;
    xRight = rhs.xRight;
    yBottom = rhs.yBottom;
    yTop = rhs.yTop;


    _dx = rhs._dx;
    _dy = rhs._dy;

    deadFrac = rhs.deadFrac;
    frameTime = rhs.frameTime;
    exposure = rhs.exposure;


    filter = rhs.filter;
    
    return *this;

};

/**********************************************************************
* This is for unrotated images
*
* NB WINDOWX0 is the window x-origin in the full-fram window (0-248)
*
* NB WINDOWY0 is the window y-origin in the full-frame window (0-248)
*
**********************************************************************/
bool WcsData2::load( string file)
{
  //     cout << "WCSData2- Loading data from " << file << endl;

    DataSet *dataSet = dataSetServer->open( file, Dal::Read);
    


    _obsId = dataSet->attribute("OBS_ID")->asString();



    if ( dataSet->hasAttribute("WINDOWX0") )
    {
   
        cdelt1 = 1.0;
        cdelt2 = 1.0;

        /*******************************************************************
	* NB WINDOWX0 will be the undistorted and unbinned x-coordinate of pixel(0, 0)
        *    WINDOWY0 will be the undistorted and unbinned y-coordinate of pixel(0, 0)
	*
	*    WINDOWDX will be the x-axis length (unbinned pixels)
     	*    WINDOWDY will be the y-axis length (unbinned pixels)
        *
        *
        ********************************************************************/

        crpix1 = dataSet->attribute("WINDOWX0")->asReal();
        crpix2 = dataSet->attribute("WINDOWY0")->asReal();


        if(dataSet->hasAttribute( "CRPIX1" ))
	{   
            crval1 = dataSet->attribute("CRPIX1")->asReal();
            crval2 = dataSet->attribute("CRPIX2")->asReal();
        }
        else
	{
            crval1 = 0.0;
            crval2 = 0.0;;
        }    

        binAxis = (int)pow(2, (double)dataSet->attribute("BINAX1")->asInt());
        binBpe = dataSet->attribute("BINBPE")->asBool();
	if( binBpe)
	{ 
	    binAxis *= 2; 
     

	}
        /************************************************************************
	*
        * The actual size of the image
        *
        *************************************************************************/
        nx = dataSet->attribute("WINDOWDX")->asInt() / binAxis;
        ny = dataSet->attribute("WINDOWDY")->asInt() / binAxis;



        xLeft = crpix1;
        xRight = xLeft +  dataSet->attribute("WINDOWDX")->asReal();
        yBottom = crpix2;
        yTop = yBottom +  dataSet->attribute("WINDOWDY")->asReal();

    }
    else
    {
 
      //        errstr << "No attribute WINDOWX0 in file  " << file << "!- terminating program\n" 
      //	       << Sas::message(Msg::App, Msg::Sparse);

        GetWindowParameters( dataSet, binAxis, binBpe, crpix1, crpix2, nx, ny);


        xLeft = crpix1;
        xRight = xLeft +  nx;
        yBottom = crpix2;
        yTop = yBottom + ny;








	//    dataSetServer->close(dataSet); // Finished with the file
	// return false;           
	     //	     Sas::fatal("Bad file", "Attribute CDELT1 does not exist in file");


    }

     _area = (xRight - xLeft) * (yTop - yBottom);




    //    cout << "Wcsdata2 load - binaxis=" << binAxis << endl;
    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = dataSet->array("PRIMARY");  
    longVec size = array->size();
  

    nx = (int)size[1];
    ny = (int)size[0];



        
    deadFrac = dataSet->attribute("DEADFRAC")->asReal();
    frameTime = dataSet->attribute("FRAMTIME")->asReal();

    // SPR-7299 : set deadfrac from the frame time
    if (frameTime>0.0)
       deadFrac = 0.174 / frameTime;
    else
       deadFrac = 0.0;

    exposure = dataSet->attribute("EXPOSURE")->asReal();
 

    filter = dataSet->attribute("FILTER")->asString();
   
 

    if (dataSet->hasAttribute("OBJECT"))
    {
        object = dataSet->attribute("OBJECT")->asString();
    }




    dataSetServer->close(dataSet); // Finished with the file
   
    //    cout << "Loaded data okay\n";

   
    return true;

}



/*************************************************************************
*
*
*************************************************************************/

/***************************************************************************
*
*
***************************************************************************/
void WcsData2::setImageParameters()
{


  //   cout << "Wcsdata2- setimageParameters\n";
    crval1 = 0.0;
    crval2 = 0.0;


    /************************************************************************
    * xLeft, xRight, yTop and yBottom are in full-frame image coordinates
    *************************************************************************/
    nx = (int)ceil( xRight - xLeft) /  binAxis; 
    ny = (int)ceil( yTop - yBottom) /  binAxis;

    

    //    cout << "nx=" << nx << " ny=" << ny << " binaxis=" << binAxis << endl;
 

}

/**********************************************************************
*
*
**********************************************************************/
void WcsData2::convert() {};


/*******************************************************************
*
* Computes the RA and Dec coordinates of each corner of the image
*
*
******************************************************************/
void WcsData2::getCornerCoords()
{

    xLeft = crpix1;
    yBottom = crpix2;

    xRight = crpix1 + double(nx) * binAxis;   // Full frame x coordinate 
    yTop =   crpix2 + double(ny) * binAxis;   // Full-frame y coordinate



}

/**************************************************************************
*
* Convert image 1 coordinates  x1, y1 on image 1 to image 2 coordinates x2, y2  
*
***************************************************************************/
inline void WcsData2::convertX1Y1ToX2Y2(double x1, double y1, OMWCS &c2, double &x2, double &y2)
{


  


    /**************************************************************
    *
    * Convert image coords x1, y1 to full frame coords x and y
    *
    ****************************************************************/


    double x = 0.0;
    double y = 0.0;

    imageToFullFrameCoords(x1, y1, crpix1, crpix2, binAxis, x, y);
 

    /*****************************************************************
    *
    * Need to remove the distortion to get the undistorted coordinates for c1
    * Then need to find the image coordinates in c2 ?????
    *
    ******************************************************************/

   

    /**************************************************************
    *
    * Convert full frame coords x, y to image coords x2, y2 
    *
    ****************************************************************/
    fullFrameToImageCoords(x, y, c2.data->crpix1, c2.data->crpix2, 
                           c2.data->binAxis, x2, y2);





   
}

/**********************************************************************
*
*
*
*
**********************************************************************/
bool WcsData2::withinImage(double x, double y)
{

    if ( x >= 0.5 && y >= 0.5 && x <= double(nx) + 0.5 &&  y <= double(ny) + 0.5)
        return true;
    else
        return false;


}
/**********************************************************************************
*
*
**********************************************************************************/
// 2014-02-02 adding max deadFrac and frameTime
void WcsData2::createOutputFile(string fileName, 
				vector< vector<real32> > &image, 
				vector< vector<real32> > &exposureImage,
				vector< vector<int16> > &qualityImage,
				//vector<string> &fileList)
				vector<string> &fileList, double deadFracMax, double frameTimeMax)



{

    if(fileList.size() == 0)
    {
        errstr << "L2024 Cannot create output image- no image files"
	       << Sas::warning("nofiles");

        return;


    }
    int numStacked = (int)fileList.size();
    
    //cout << "L2033 numStacked=" << numStacked << " frameTimeMax=" << frameTimeMax << endl;

    DataSet * set  = dataSetServer->open(fileName, Dal::Create);

  
    int nx = (int)image.size();
    int ny = (int)image[0].size();
   

    /*********************************************************
    *
    *********************************************************/
    normalizeImage(image, exposureImage);



    addBlock(set, "PRIMARY", image, 0);
    addBlock(set, "EXPOSURE", exposureImage, 1);

    addBlock(set, "QUALITY", qualityImage, 2);


    /*******************************************************************
    *
    *
    *******************************************************************/
    *set->addAttribute("DATAMODE", "") = "IMAGING";

    *set->addAttribute("FILTER", "") = filter;
  
    *set->addAttribute("OBJECT", "") = object;

    *set->addAttribute("OBS_ID", "") = _obsId;


    *set->addAttribute("DATE-OBS", "") = startObs;
    *set->addAttribute("DATE-END", "") = endObs;

 

    /*****************************************************************
    *
    * Determine BINAX1 from plate scale
    *
    ******************************************************************/
  

    binAxis /= 2;



    *set->addAttribute("BINAX1", "") = binAxis;
    *set->addAttribute("BINAX2", "") = binAxis;
    *set->addAttribute("BINBPE", "") = binBpe;

    *set->addAttribute("EXPOSURE", "") = 1;

    //*set->addAttribute("DEADFRAC", "") = deadFrac;
    //*set->addAttribute("FRAMTIME", "") = frameTime;

    *set->addAttribute("DEADFRAC", "") = deadFracMax;
    *set->addAttribute("FRAMTIME", "") = frameTimeMax;

    *set->addAttribute("WINDOWX0", "") = crpix1;
    *set->addAttribute("WINDOWDX", "") = nx;
    *set->addAttribute("WINDOWY0", "") = crpix2;
    *set->addAttribute("WINDOWDY", "") = ny;


    *set->addAttribute( "NSTACKED", "Number of images stacked") = numStacked;


    // for(unsigned i = 0; i < fileList.size(); i++)
    // {
    //  char text[100];
    //  sprintf(text, "IMAGE%03d", i + 1 );
    //  *set->addAttribute( text, "Image stacked") = fileList[i];

    // }


   
    dataSetServer->close(set); // Finished with the file
    


}
/*************************************************************
*
*
*
***************************************************************/
void WcsData2::correctReferencePixel(double dx, double dy)
{



  crpix1 -= dx;
  crpix2 -= dy;

   


}

/*************************************************************
*
* Corrects the reference pixel crpix1 and crpix2
* ie remove the offsets
* NB - the stored offsets will be in unbinned pixels
***************************************************************/
void WcsData2::correctReferencePixel()
{


  
    

    crpix1 -= _dx;
    crpix2 -= _dy;






}
/*******************************************************************
*
*
*
*
********************************************************************/
void WcsData2::updateReferencePixelCorrection(OMWCS &c, double dx, double dy)
{

   


    /***************************************************************
    *
    * All offsets (dx and dy) will be in unbinned pixels
    *
    ****************************************************************/
    if(c.data->binAxis  == 1)
    {
        if(binAxis  == 1)
        {
            _dx += c.data->_dx + dx;
            _dy += c.data->_dy + dy;
        } 
        else
        {
            _dx += c.data->_dx + dx * 2.0;
            _dy += c.data->_dy + dy * 2.0;
        } 
    }
    else
    {

        if(binAxis  == 1)
        {
            _dx += c.data->_dx + dx;
            _dy += c.data->_dy + dy;
        } 
        else
        {
            _dx += c.data->_dx + dx * 2.0;
            _dy += c.data->_dy + dy * 2.0;
        } 

      
    } 
 

}


/*******************************************************************************
*
*
*
********************************************************************************/
double WcsData2::distance(double x, double y, OMWCS &c, double &x2, double &y2)
{


    double x1 = 0.0, y1 = 0.0; 
    convertX1Y1ToX2Y2(x, y, c, x1,  y1);


    double dx = x1 - x2;
    double dy = y1 - y2;
    return hypot(dx, dy);

    
}

/*******************************************************************************
*
*
********************************************************************************/
double WcsData2::overlapFraction(OMWCS &c2)
{

  //cout << "Overlap fraction line 1495\n";
  //cout << "xleft=" << xLeft << " right=" << xRight << endl;
  // cout << "ybottom=" << yBottom << " top=" << yTop << endl;

  // cout << "c2 values\n";
  // cout << "xleft=" << c2.data->xLeft << " right=" << c2.data->xRight << endl;
  // cout << "ybottom=" << c2.data->yBottom << " top=" << c2.data->yTop << endl;


  double x1 = 0.0, x2 = 0.0, y1 = 0.0, y2 = 0.0;
  x1 = max(xLeft,  c2.data->xLeft);
  x2 = min(xRight,  c2.data->xRight);

  y1 = max(yBottom,  c2.data->yBottom);
  y2 = min(yTop,  c2.data->yTop);




  //cout << "Line 1528 x1=" << x1 << " x2=" << x2 << " y1=" << y1 << " y2=" << y2 << endl;
  if( (x2 - x1) < 1.0 || (y2 - y1) < 1.0) return 0.0;

  double area1 = (x2 - x1) * (y2 - y1);
  double area2 = ((xRight - xLeft) * (yTop - yBottom)); 

  double area3 = ((c2.data->xRight - c2.data->xLeft) * (c2.data->yTop - c2.data->yBottom));

  //  cout << "area1=" << area1 << " area2=" << area2 << " area3=" << area3 << endl;
  return area1 / min(area2, area3); 



}



/*************************************************************
*
* This is for unrotated images
*
*
**************************************************************/
void WcsData2::convertToLowResImage(vector< vector<float> > &image, 
                vector< vector<float> > &exposureImage,
		vector< vector<int16> > &qualityImage)

{
    if(binAxis == 2) return;


    int nx = 0, ny = 0;
    if ( (image.size() / 2) * 2 == image.size() )
      nx = image.size() / 2;
    else
      nx = image.size() / 2 + 1;


    if ( (image[0].size() / 2) * 2 == image[0].size() )
      ny = image[0].size() / 2;
    else
      ny = image[0].size() / 2 + 1;

    vector< vector<float> > tempImage(nx, vector<float>(ny, -9999.9));
    vector< vector<float> > tempExposureImage(nx, vector<float>(ny, -9999.9));

    vector< vector<int16> > tempQualityImage(nx, vector<int16>(ny, 0));


    --nx;
    --ny;

    for (int i = 0; i < (int)image.size(); i+= 2)
    {

        int i1 = i / 2;

        for (int j = 0; j < (int)image[0].size(); j+= 2)
        {
  
 
            int j1 = j / 2;
            int count = 0;
            double sum1 = 0.0, sum2 = 0.0;

            int16 qFlag;
            for(int k = i; k <= min(i + 1, nx); k++)
	    {
                for(int l = j; l <= min(j + 1, ny); l++)
		{
                  if (!isNull(image[k][l] ) )
		  {
		      sum1 += (double)image[k][l];
                      sum2 += (double)exposureImage[k][l];
                      ++count;
                
                      for(int n = 0; n < 12; n++)
		      {
		          if (bitSet( qualityImage[k][l], n ) )
		          {
			    setBit(qFlag, n);
                          }
		      }

		  }


		} 

            }  

            if(count > 0)
	    {
	      tempImage[i1][j1] = sum1;
              tempExposureImage[i1][j1] = sum2 / (double)count;
              tempQualityImage[i1][j1] = qFlag; 
	    } 
	}
    }

    image = tempImage;
    exposureImage = tempExposureImage;
    qualityImage = tempQualityImage;


    /************************************************************
    *
    *
    *
    *
    **************************************************************/
    binAxis = 2;
    nx = nx;
    ny = ny;

}



/*******************************************************************
*
*
*
*******************************************************************/
void GetWindowParameters(DataSet *dataSetH, int &binAxis, bool &binBpe, 
         double &windowX0, double &windowY0, int &nx, int &ny)
{    


    /*********************************************************************
    *
    * Get the binning factor
    *
    *********************************************************************/
    binAxis = (int)pow(2, (double)dataSetH->attribute("BINAX1")->asInt());
    binBpe = dataSetH->attribute("BINBPE")->asBool();

    nx = ny = 0;

    if( binBpe)
    { 
        binAxis *= 2; 
     
    }

    /****************************************************
    *
    *
    ****************************************************/
    if ( dataSetH->hasAttribute("DETSEC"))
    {

      // cout << "Line 2396\n";
        string windowString = dataSetH->attribute( "DETSEC" )->asString();

      
        int windowX1, windowY1, x0, y0;
        char c;
        sscanf(windowString.c_str(), "%c %4d %c %4d %c %4d %c %4d", &c, &x0, &c, &windowX1, &c, &y0, &c, &windowY1);

	//        cout << "windowstring=" << windowString << endl;
	//	cout << "Line 2404 x0=" << windowX0 << endl;
	//	cout << "Line 2405 y0=" << windowY0 << endl;

	//	cout << "Line 2404 windowX1=" << windowX1 << endl;
	//	cout << "Line 2404 windowY1=" << windowY1 << endl;

    
        nx = windowX1 - x0 + 1;  
        ny = windowY1 - y0 + 1;
        windowX0 = (double)x0;
        windowY0 = (double)y0;


    }
    if (binBpe) 
    {
        binAxis *= 2;     
        windowX0 *= binAxis;
	windowY0 *= binAxis;
    }

}












